
if ( !window.requestAnimationFrame ) {
    window.requestAnimationFrame = ( function() {
        return window.webkitRequestAnimationFrame ||
            window.mozRequestAnimationFrame ||
            window.oRequestAnimationFrame ||
            window.msRequestAnimationFrame;
    } )();
};

function Renderer()

{

    var that = this;
    that.geoList = [];
    that.bufferViews = {};
    that.header;
    that.ibufferList = [];
    that.vbufferList = [];
    that.ibufferLength = 0; //dependent on container type
    that.vbufferLength = 0; //dependent on container type
    var CANVAS = document.getElementById("myCanvas");

    CANVAS.width  = 1024;
    CANVAS.height = 768;

    // WEBGL CONTEXT

    try
    {
        var gl = CANVAS.getContext("experimental-webgl");
    }
    catch(e)
    {
        console.log("no webgl support");
        return false;
    }
    var ext = (
        gl.getExtension("WEBGL_compressed_texture_s3tc") ||
        gl.getExtension("MOZ_WEBGL_compressed_texture_s3tc") ||
        gl.getExtension("WEBKIT_WEBGL_compressed_texture_s3tc")
        );
    log(ext);
    that.tex = gl.createTexture();


    // SHADERS

    var shader_vertex_source   = document.getElementById("popVS").textContent;
    var shader_fragment_source = document.getElementById("popFS").textContent;

    function get_shader(source, type, typeString)
    {
        var shader = gl.createShader(type);
        gl.shaderSource(shader, source);
        gl.compileShader(shader);
        if(!gl.getShaderParameter(shader, gl.COMPILE_STATUS))
        {
            console.log("error in "+typeString+ " shader: " + gl.getShaderInfoLog(shader));
        }
        return shader;
    }

    var shader_vertex   = get_shader(shader_vertex_source, gl.VERTEX_SHADER, "VERTEX");
    var shader_fragment = get_shader(shader_fragment_source, gl.FRAGMENT_SHADER, "FRAGMENT");

    var shader_program = gl.createProgram();
    gl.attachShader(shader_program, shader_vertex);
    gl.attachShader(shader_program, shader_fragment);

    gl.linkProgram(shader_program);

    var shpos = {};

    var _Pmatrix = gl.getUniformLocation(shader_program, "Pmatrix");
    var _Vmatrix = gl.getUniformLocation(shader_program, "Vmatrix");
    var _Mmatrix = gl.getUniformLocation(shader_program, "Mmatrix");
    shpos.decodeScale_position = gl.getUniformLocation(shader_program, "decodeScale_position");
    shpos.decodeOffset_position = gl.getUniformLocation(shader_program, "decodeOffset_position");
    shpos.decodeScale_normal = gl.getUniformLocation(shader_program, "decodeScale_normal");
    shpos.decodeOffset_normal = gl.getUniformLocation(shader_program, "decodeOffset_normal");
    shpos.decodeScale_texcoord = gl.getUniformLocation(shader_program, "decodeScale_texcoord");
    shpos.decodeOffset_texcoord = gl.getUniformLocation(shader_program, "decodeOffset_texcoord");
    shpos.popLevel = gl.getUniformLocation(shader_program, "popLevel");
    shpos.bbMinModF = gl.getUniformLocation(shader_program, "bbMinModF");
    shpos.bbMaxModF = gl.getUniformLocation(shader_program, "bbMaxModF");

    shpos.position = gl.getAttribLocation(shader_program, "position");
    shpos.color    = gl.getAttribLocation(shader_program, "color");
    shpos.normal   = gl.getAttribLocation(shader_program, "normal");
    shpos.texcoord = gl.getAttribLocation(shader_program, "texcoord");

    gl.enableVertexAttribArray(shpos.position);
    //gl.enableVertexAttribArray(shpos.color);
    gl.enableVertexAttribArray(shpos.normal);
    gl.enableVertexAttribArray(shpos.texcoord);

    gl.useProgram(shader_program);

    var projMatrix  = libs.get_projectionMat(40, CANVAS.width / CANVAS.height, 1, 100);
    var viewMatrix = libs.get_identityMat();
    libs.translateZ(viewMatrix, -15);
    libs.rotateX(viewMatrix, 0.35);
    libs.translateX(viewMatrix, -8);
    libs.translateY(viewMatrix, -150);
    var modelMatrix = libs.get_identityMat();

    var pressedKeys = {};
    this.popLevel = 16;

    this.handleKeyDown = function(event)
    {
        console.log(event.keyCode);
        pressedKeys[event.keyCode] = true;
        if(event.keyCode == 51){
            that.popLevel = that.popLevel -1;
        }
        if(event.keyCode == 52){
            that.popLevel = that.popLevel +1;
        }
    }
    this.handleKeyUp = function(event)
    {
        pressedKeys[event.keyCode] = false;
    }

    document.onkeydown = that.handleKeyDown;
    document.onkeyup = that.handleKeyUp;

    this.handleKeys = function()
    {
        if(pressedKeys[87] || pressedKeys[38]){
            this.moveForward();
        }
        if(pressedKeys[83] || pressedKeys[40]){
            this.moveBackward();
        }
        if(pressedKeys[82]){
            this.moveForward(20);
        }
        if(pressedKeys[70]){
            this.moveBackward(20);
        }
        if(pressedKeys[65]){
            this.moveLeft();
        }
        if(pressedKeys[68]){
            this.moveRight();
        }
        if(pressedKeys[81]){
            this.moveUp();
        }
        if(pressedKeys[69]){
            this.moveDown();
        }
        if(pressedKeys[72]){
            this.rotateLeft();
        }
        if(pressedKeys[71]){
            this.rotateDown();
        }
        if(pressedKeys[84]){
            this.rotateUp();
        }
        if(pressedKeys[74]){
            this.rotateRight();
        }

    }

    this.moveForward = function(distance)
    {
        if(!distance) distance = 5;
        libs.translateZ(viewMatrix, distance);
    }

    this.moveBackward = function(distance)
    {
        if(!distance) distance = 5;
        libs.translateZ(viewMatrix, -distance);
    }

    this.moveUp = function(distance)
    {
        if(!distance) distance = 5;
        libs.translateY(viewMatrix, -distance);
    }

    this.moveDown = function(distance)
    {
        if(!distance) distance = 5;
        libs.translateY(viewMatrix, distance);
    }

    this.moveLeft = function(distance)
    {
        if(!distance) distance = 5;
        libs.translateX(viewMatrix, distance);
    }

    this.moveRight = function(distance)
    {
        if(!distance) distance = 5;
        libs.translateX(viewMatrix, -distance);
    }

    this.rotateUp = function(distance)
    {
        if(!distance) distance = .1;
        libs.rotateX(viewMatrix, -distance);
    }

    this.rotateRight = function(distance)
    {
        if(!distance) distance = .1;
        libs.rotateY(viewMatrix, distance);
    }

    this.rotateDown = function(distance)
    {
        if(!distance) distance = .1;
        libs.rotateX(viewMatrix, distance);
    }

    this.rotateLeft = function(distance)
    {
        if(!distance) distance = .1;
        libs.rotateY(viewMatrix, -distance);
    }

    this.tick = function()
    {
        that.handleKeys();
        if(that.header)that.draw();
        simProg();
        requestAnimationFrame(that.tick);
    }

    this.isIndexBufferView = function(header, bufferViewKey, callback)
    {
        if(Object.keys(header.accessors.indexViews).length <= 0)
        {
            header.bufferViews[bufferViewKey].isIndexBufferView = isIndexBufferView;
            callback();
        }
        var count = Object.keys(header.accessors.indexViews).length;
        var isIndexBufferView = false;
        Object.keys(header.accessors.indexViews).forEach(function(indexViewKey){
            if(header.accessors.indexViews[indexViewKey].bufferView == bufferViewKey)
            {
                isIndexBufferView = true;
            }
            if(--count <= 0)
            {
                header.bufferViews[bufferViewKey].isIndexBufferView = isIndexBufferView;
                callback();
            }
        });
    }

    this.createBuffers = function(header)
    {
        that.header = header;
        Object.keys(header.bufferViews).forEach(function(bufferViewKey){
            that.isIndexBufferView(header, bufferViewKey, function(){
                header.bufferViews[bufferViewKey].progress = 0;
                header.bufferViews[bufferViewKey].glBuffer = gl.createBuffer();
                var target = header.bufferViews[bufferViewKey].isIndexBufferView?gl.ELEMENT_ARRAY_BUFFER:gl.ARRAY_BUFFER;
                gl.bindBuffer(target, header.bufferViews[bufferViewKey].glBuffer);
                gl.bufferData(target, header.bufferViews[bufferViewKey].byteLength, gl.STATIC_DRAW);
            });
        });
    }

    this.updateProgression = function()
    {
        for(var k = 0; k < Object.keys(that.header.meshes).length; k++){
            var mesh = that.header.meshes[Object.keys(that.header.meshes)[k]];
            var maxProgression = mesh.meta.attributeProgression.length - 1;
            for(var j = 0; j < Object.keys(mesh.attributes).length; j++){
                var attributeView = that.header.accessors.attributeViews[mesh.attributes[Object.keys(mesh.attributes)[j]]];
                var attributeBufferView = that.header.bufferViews[attributeView.bufferView];
                for (var i = 0; i <= maxProgression; i++)
                {
                    var bytesNeeded = (mesh.meta.attributeProgression[i] - 1) * attributeView.byteStride + that.getComponentCount(attributeView.type) * that.getComponentSize(attributeView.componentType);
                    if(attributeBufferView.progress < (attributeView.byteOffset + bytesNeeded))
                    {
                        console.log("attribute progression not reached " + mesh.meta.attributeProgression[i] + " " + attributeBufferView.progress + " " + bytesNeeded + " " + attributeView.byteOffset + " " + i);
                        maxProgression = i - 1;
                        break;
                    }
                }
            }
            if(mesh.indices)
            {
                var indexView = that.header.accessors.indexViews[mesh.indices];
                var indexBufferView = that.header.bufferViews[indexView.bufferView];
                for (var i = 0; i <= maxProgression; i++)
                {
                    var bytesNeeded = mesh.meta.indexProgression[i] * 2; //16Bit indices
                    if(indexBufferView.progress < indexView.byteOffset + bytesNeeded)
                    {
                        console.log("index progression not reached " + mesh.meta.indexProgression[i] + " " + indexBufferView.progress + " " + bytesNeeded)
                        maxProgression = i - 1;
                        break;
                    }
                }
                mesh.elementsToDraw = mesh.meta.indexProgression[maxProgression];
            }
            else
            {
                mesh.elementsToDraw = mesh.meta.attributeProgression[maxProgression];
            }
            mesh.maxProgression = maxProgression;
        }

    }

    this.addChunk = function(chunkName, chunk)
    {
        updateOSDChunks(chunkName);
        //check for geometry
        for(var i = 0; i < Object.keys(that.header.bufferViews).length; i++){
            var bufferView = that.header.bufferViews[Object.keys(that.header.bufferViews)[i]];
            var offset = 0;
            for (var j = 0; j < bufferView.chunks.length; j++)//search all chunk refs in bufferViews
            {
                if(bufferView.chunks[j] == chunkName) //if chunk found in buffer
                {
                    //add chunk to buffer
                    bufferView.progress += that.header.bufferChunks[chunkName].byteLength;
                    console.log("add chunk " +that.header.bufferChunks[chunkName].byteLength + " " + chunk.byteLength + " " + Object.keys(that.header.bufferViews)[i]);
                    var target = (bufferView.isIndexBufferView)? gl.ELEMENT_ARRAY_BUFFER : gl.ARRAY_BUFFER;

                    gl.bindBuffer(target, bufferView.glBuffer);
                    gl.bufferSubData(target, offset, chunk);

                    break;
                }
                offset += that.header.bufferChunks[bufferView.chunks[j]].byteLength;
            }
        }

        //check for texture
        for(var i = 0; i < Object.keys(that.header.textureViews).length; i++){
            var textureViewKey = Object.keys(that.header.textureViews)[i];
            var textureView = that.header.textureViews[textureViewKey];
            var offset = 0;
            for (var j = 0; j < textureView.chunks.length; j++)//search all chunk refs in bufferViews
            {
                if(textureView.chunks[j] == chunkName) //if chunk found in buffer
                {
                    if(!textureView.progress)
                    {
                        textureView.progress = 0;
                    }
                    if(!textureView.byteOffset)
                    {
                        textureView.byteOffset = [];
                    }
                    textureView.progress += that.header.bufferChunks[chunkName].byteLength;
                    textureView.byteOffset.push(chunk.byteOffset);
                    console.log("add texchunk " +that.header.bufferChunks[chunkName].byteLength + " " + chunk.byteLength + " " + Object.keys(that.header.bufferViews)[i]);
                    for(var k = 0; k < Object.keys(that.header.textures).length; k++){
                        var texture = that.header.textures[Object.keys(that.header.textures)[k]];
                        if(texture.textureView == textureViewKey)
                        {
                            if(!texture.uploadedImages)
                            {
                                texture.uploadedImages = 0;
                            }
                            if(!textureView.uploadedBytes)
                        {
                            textureView.uploadedBytes = 0;
                        }
                            var bytesNeeded = textureView.uploadedBytes;
                            for (var level = texture.uploadedImages; level < texture.imageByteLengths.length; level++)
                            {
                                bytesNeeded += texture.imageByteLengths[level];
                                if (textureView.progress >= bytesNeeded)
                                {
                                    that.uploadTex(texture, chunk.buffer, level);
                                    texture.uploadedImages++;
                                    textureView.uploadedBytes += texture.imageByteLengths[level];
                                }
                                else
                                {
                                    break;
                                }
                            }
                        }
                    }

                    break;
                }
                offset += that.header.bufferChunks[textureView.chunks[j]].byteLength;
            }
        }
        that.updateProgression();
//        requestAnimationFrame(that.draw);
    }

    this.uploadTex = function(texture, arrayBuffer, level)
    {
        var textureView = that.header.textureViews[texture.textureView];
        var byteOffset = textureView.byteOffset[level];
        var byteLength = texture.imageByteLengths[level];


        log(byteOffset + " " + byteLength);
        var arrayBufferView = new Uint8Array( arrayBuffer, byteOffset, byteLength );

        var blob = new Blob([arrayBufferView], {type: 'image/png'});
//        var img = new Image();
        var img =  document.getElementById("tex");
        img.onload = function(){
            if(!that.tex.images)
            {
                that.tex.images = [];
            }
            that.tex.images.push(img);

            gl.bindTexture(gl.TEXTURE_2D, that.tex);
//            for (var i = 0; i < that.tex.images.length; i++)
//            {
//                gl.texImage2D(gl.TEXTURE_2D, i, texture.internalFormat, texture.format, texture.type, that.tex.images[that.tex.images.length - 1 - i]);
//            }

            gl.texImage2D(gl.TEXTURE_2D, 0, texture.internalFormat, texture.format, texture.type, img);
//            gl.generateMipmap(gl.TEXTURE_2D);

            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
//            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST_MIPMAP_NEAREST);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
            gl.bindTexture(gl.TEXTURE_2D,null);
//            requestAnimationFrame(that.draw);
        }
        var urlCreator = window.URL || window.webkitURL;
        var imageUrl = urlCreator.createObjectURL( blob );
        img.src = imageUrl;
    }

    this.loadBuffer = function(mesh)
    {
        if(mesh.indices)
        {
            var indexBufferView = that.header.accessors.indexViews[mesh.indices].bufferView;
            gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, that.header.bufferViews[indexBufferView].glBuffer);
        }
        var attribBufferView = that.header.accessors.attributeViews[mesh.attributes["position"]].bufferView;
        gl.bindBuffer(gl.ARRAY_BUFFER, that.header.bufferViews[attribBufferView].glBuffer);
    }

    this.getComponentCount = function(type)
    {
        if(type == "SCALAR")
        {
            return 1;
        }
        if(type == "VEC2")
        {
            return 2;
        }
        if(type == "VEC3")
        {
            return 3;
        }
    }

    this.getComponentSize = function(componentType)
    {
        if(componentType == 0x1400 || componentType == 0x1401)
        {
            return 1;
        }
        if(componentType == 0x1402 || componentType == 0x1403)
        {
            return 2;
        }
        if(componentType == 0x1404 || componentType == 0x1405 || componentType == 0x1406)
        {
            return 4;
        }
    }

    this.draw = function()
    {


        gl.uniformMatrix4fv(_Vmatrix, false, viewMatrix);

        gl.clearColor(0.0, 0.0, 0.0, 0.0);

        gl.enable(gl.DEPTH_TEST);
        gl.depthFunc(gl.LEQUAL);
        gl.clearDepth(1.0);
        gl.viewport(0.0, 0.0, CANVAS.width, CANVAS.height);
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
        var projMatrix = libs.get_projectionMat(40, CANVAS.width / CANVAS.height, 1, 500000);
        gl.uniformMatrix4fv(_Pmatrix, false, projMatrix);

        var drawId = 0;
        Object.keys(that.header.meshes).forEach(function(meshKey){
            var mesh = that.header.meshes[meshKey];
            that.drawCall(drawId++, mesh);
        });

        gl.flush();

    }

    this.currentProgression = 0;
    this.drawCall = function(drawId, mesh)
    {
        if(mesh.maxProgression == undefined || mesh.maxProgression < 0)
        {
            return;
        }
        //BIND BUFFERS

//        console.log("draw elements: "+ mesh.elementsToDraw);
        that.loadBuffer(mesh);

        // MATRIX


        modelMatrix = libs.get_identityMat();
//        libs.translateX(modelMatrix, 3 * drawId);
        libs.rotateX(modelMatrix, -1.57);
        libs.rotateY(modelMatrix, -1.1);
        libs.translateZ(modelMatrix, -500);
        libs.translateY(modelMatrix, -100);
        libs.translateX(modelMatrix, 100);
        //libs.rotateX(modelMatrix, 0.3);

        // UNIFORMS
        gl.uniformMatrix4fv(_Mmatrix, false, modelMatrix);
//        console.log(that.header.accessors.attributeViews[mesh.attributes.position].decodeOffset[0]);
//        var test = [1.0, 1.0, 0.5];
//        gl.uniform3fv(shpos.decodeOffset_position, test);
        gl.uniform3fv(shpos.decodeOffset_position, that.header.accessors.attributeViews[mesh.attributes.position].decodeOffset);
        gl.uniform3fv(shpos.decodeScale_position, that.header.accessors.attributeViews[mesh.attributes.position].decodeScale);
        gl.uniform3fv(shpos.decodeOffset_normal, that.header.accessors.attributeViews[mesh.attributes.normal].decodeOffset);
        gl.uniform3fv(shpos.decodeScale_normal, that.header.accessors.attributeViews[mesh.attributes.normal].decodeScale);
        gl.uniform2fv(shpos.decodeOffset_texcoord, that.header.accessors.attributeViews[mesh.attributes.texcoord].decodeOffset);
        gl.uniform2fv(shpos.decodeScale_texcoord, that.header.accessors.attributeViews[mesh.attributes.texcoord].decodeScale);
        gl.uniform1f(shpos.popLevel, Math.min(this.popLevel, mesh.maxProgression));
        gl.uniform3fv(shpos.bbMinModF, mesh.meta.bbMinModF);
        gl.uniform3fv(shpos.bbMaxModF, mesh.meta.bbMaxModF);

        if(this.currentProgression != mesh.maxProgression)
        {
            this.currentProgression = mesh.maxProgression;
            log("Progression: "  + this.currentProgression);
        }
        // DRAWING


        // TEXTURE
        if(that.tex.images && that.tex.images.length > 0)
        {
            gl.activeTexture(gl.TEXTURE0);
            gl.bindTexture(gl.TEXTURE_2D, that.tex);
            gl.uniform1i(gl.getUniformLocation(shader_program, "uSampler"), 0);
        }


        // draw
        //console.log(mesh.indices);
        for(var i = 0; i < Object.keys(mesh.attributes).length; i++){
            var attributeViewKey = mesh.attributes[Object.keys(mesh.attributes)[i]];
            var attributeView = that.header.accessors.attributeViews[attributeViewKey];
            gl.vertexAttribPointer(shpos[Object.keys(mesh.attributes)[i].toLowerCase()], that.getComponentCount(attributeView.type), attributeView.componentType, false, attributeView.byteStride, attributeView.byteOffset);
//            console.log(shpos[Object.keys(mesh.attributes)[i].toLowerCase()] + "  " + Object.keys(mesh.attributes)[i].toLowerCase() + " " + attributeView.type +  " " + attributeView.byteStride + " " + attributeView.byteOffset);
        }

        //gl.drawElements(gl.TRIANGLES, 3*12, gl.UNSIGNED_SHORT, 0);
        if(mesh.indices)
        {
            var indexView = that.header.accessors.indexViews[mesh.indices];
            gl.drawElements(mesh.primitive, mesh.elementsToDraw, indexView.componentType, indexView.byteOffset);
        }
        else
        {
            gl.drawArrays(4, 0, mesh.elementsToDraw);
        }

    }

}